# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
import sys

# Third party libs
import email.message
sys.path.insert(0, './imapclient/imapclient')  # TODO this is ugly, improve it


class Mail(dict):
    mail_native = email.message.Message()

    def __init__(self, logger, uid, mail=None):
        self.logger = logger
        self.uid = uid
        self.mail_native = mail
        self.parse_email_object()

    def parse_email_object(self):
        fields_to_store = {  # TODO store as much as possible, get rid of this static list
            'content-type': None,
            'date': None,
            'delivered-to': {'multiple': True},
            'from': None,
            'list-id': None,
            'message-id': None,
            'received': {'multiple': True, 'split': True},
            'return-path': None,
            'subject': None,
            'to': None,
            'user-agent': None,
            'x-redmine-project': None,
            'x-redmine-host': None,
            'x-gitlab-project': None,
        }

        for field, properties in fields_to_store.items():
            if type(properties) is dict:
                if properties.get('multiple', None):
                    fields = self.mail_native.get_all(field)
                    if fields is None:
                        self.logger.debug('Unable to parse raw mail with uid=%s: %s', self.uid, self.mail_native)
                        raise ValueError('Unable to parse raw mail with uid=%s' % (self.uid))

                    if properties.get('split', None):
                        _fields = []
                        for f in fields:
                            _fields.append(f.split('\r\n'))
                        _fields = fields
                    self[field] = fields
                else:
                    self[field] = self.mail_native.get(field)
            else:
                self[field] = self.mail_native.get(field)



# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
Notes:

* before calling a private function, always issue a new SELECT command
"""
import sys

# Third party libs
import email.message
sys.path.insert(0, './imapclient/imapclient')  # TODO this is ugly, improve it
from imapclient import IMAPClient

from mail import Mail


class IMAP(object):
    def __init__(self, logger, username, password, server='localhost', port=143, ssl=None, startls=None, test=False):
        self.logger = logger
        self.username = username
        self.password = password
        self.server = server
        self.port = port
        self.ssl = ssl
        self.starttls = None
        self.test = test

    def connect(self):
        self.logger.debug('Establishing IMAP TLS connection to %s and logging in with user %s', self.server, self.username)

        if self.port == 143:
            self.starttls = True
        else:
            self.starttls = None
            self.ssl = True

        self.conn = IMAPClient(host=self.server, port=self.port, use_uid=True, starttls=self.starttls, ssl=self.ssl)
        self.conn.login(self.username, self.password)

    def process_error(self, exception):
        self.logger.error('Catching IMAP exception: %s', exception)

    def select_mailbox(self, mailbox):
        self.logger.debug('Switching to mailbox %s', mailbox)
        try:
            result = self.conn.select_folder(mailbox)
        except IMAPClient.Error as e:
            self.process_error(e)
            result = None
        return result

    def search_mails(self, mailbox, criteria='ALL'):
        self.logger.debug('Searching for mails in mailbox %s and criteria=\'%s\'', mailbox, criteria)
        try:
            result = self.select_mailbox(mailbox)
            result = self.conn.search(criteria=criteria)
            mail_uids = result
        except IMAPClient.Error as e:
            self.process_error(e)
            mail_uids = []
        return list(mail_uids)

    def fetch_raw_mails(self, uids, mailbox, return_fields=[b'RFC822']):
        if len(uids) == 0:
            return []
        self.logger.debug('Fetching raw mails with uids %s', uids)

        try:
            mails = {}
            for uid in uids:
                result = self.select_mailbox(mailbox)
                result = self.conn.fetch(uid, return_fields)
                for fetch_uid, fetch_mail in result.items():
                    mails[fetch_uid] = fetch_mail
        except IMAPClient.Error as e:
            self.process_error(e)
        return mails

    def fetch_mails(self, uids, mailbox):
        if type(uids) is not list:
            uids = [uids]
        if len(uids) == 0:
            return {}
        self.logger.debug('Fetching mails with uids %s', uids)

        raw_mails = self.fetch_raw_mails(uids, mailbox)
        mails = {}
        for raw_uid, raw_mail in raw_mails.items():
            mail = Mail(logger=self.logger, uid=raw_uid, mail=email.message_from_bytes(raw_mails[raw_uid][b'RFC822']))
            mails[raw_uid] = mail
        return mails

    def move_mail(self, mail, source, destination, delete_old=True, expunge=True, set_flags=None):
        if self.test:
            self.logger.info('Would have moved mail message-id="%s" from "%s" to "%s", skipping because of beeing in testmode',
                             mail.get('message-id'), source, destination)
        else:
            self.select_mailbox(source)
            self._move_mail(mail, source, destination, delete_old, expunge, set_flags)

    def set_mailflags(self, uids, mailbox, flags=[]):
        if self.test:
            self.logger.info('Would have set mail flags on message uids "%s"', str(uids))
        else:
            self.select_mailbox(mailbox)
            return self._set_mailflags(uids, flags)

    def _move_mail(self, mail, source, destination, delete_old=True, expunge=True, set_flags=None):
        self.logger.info('Moving mail message-id="%s" from "%s" to "%s"', mail.get('message-id'), source, destination)
        result = self.search_mails(source, criteria='HEADER MESSAGE-ID "{0}"'.format(mail.get('message-id')))
        uid = result[0]

        self._copy_mail(uid, destination)

        if delete_old:
            self._delete_mails(uid)

            if expunge:
                self._expunge()

        if type(set_flags) is list:
            self.select_mailbox(destination)
            uids = self.search_mails(destination, criteria='HEADER MESSAGE-ID "{0}"'.format(mail.get('message-id')))
            self._set_mailflags(uids, set_flags)

    def _set_mailflags(self, uids, flags=[]):
        self.logger.debug('Setting flags=%s on mails uid=%s', flags, uids)
        try:
            result = self.conn.set_flags(uids, flags)
        except IMAPClient.Error as e:
            self.process_error(e)
            result = None
        return result

    def _expunge(self):
        self.logger.info('Expunge mails')
        try:
            result = self.conn.expunge()
        except IMAPClient.Error as e:
            self.process_error(e)
            result = None
        return result

    def _create_mailbox(self, mailbox):
        self.logger.info('Creating mailbox %s', mailbox)
        try:
            result = self.conn.create_folder(mailbox)
        except IMAPClient.Error as e:
            self.process_error(e)
            result = None
        return result

    def _mailbox_exists(self, mailbox):
        self.logger.debug('Checking wether mailbox %s exists', mailbox)
        try:
            result = self.conn.folder_exists(mailbox)
        except IMAPClient.Error as e:
            self.process_error(e)
            result = None
        return result

    def _delete_mails(self, uids):
        self.logger.info('Deleting mails uid="%s"', uids)
        try:
            result = self.conn.delete_messages(uids)
        except IMAPClient.Error as e:
            self.process_error(e)
            result = None
        return result

    def _copy_mail(self, uids, destination):
        self.logger.info('Copying mails uid="%s" to "%s"', uids, destination)
        try:
            result = self.conn.copy(uids, destination)
        except IMAPClient.Error as e:
            self.process_error(e)
            if not self._mailbox_exists(destination):
                self.logger.info('Mailbox %s doesn\'t even exist! Creating it for you now', destination)
                result = self._create_mailbox(destination)
                result = self._copy_mail(uids, destination)
            else:
                result = None
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""tabellarius

Loglevels in use:
* DEBUG
* ERROR
* CRITICAL
* INFO

TODO:
* TLS communication is UNCONFIGURED (using library defaults)
* support IDLE connections
* improve error handling, throw excetions?
* filtering fields like deliverd-to/ received/ body are not supported yet
* manage different namespaces
* expunge at the end only
"""
# Third party libs
import argparse

from imap import IMAP
from ruleset import RuleSet
from misc import ConfigParser, Helper


def main():
    version = '0.0.1'
    program_name = 'tabellarius'
    parser = argparse.ArgumentParser(prog=program_name, description='A mail-sorting tool that is less annoying')

    # General args
    parser.add_argument('-V', action='version', version='%(prog)s {version}'.format(version=version))
    parser.add_argument('-t', '--test',
                        action='store_true',
                        dest='test',
                        help='Run in test mode, run read-only IMAP commands only',
                        default=None)
    parser.add_argument('--confdir',
                        action='store',
                        dest='confdir',
                        help='directory to search for configuration files (default: config/)',
                        default='config/')

    parser_results = parser.parse_args()
    confdir = parser_results.confdir
    test = parser_results.test

    # Config Parsing
    cfg_parser = ConfigParser(confdir)
    config = cfg_parser.dump()
    if test is not None:
        config['settings']['test'] = test

    # Logging
    logconfig = config.get('settings', {}).get('logging', {})
    logger = Helper().create_logger(program_name, logconfig)

    # Let's start working now
    logger.debug('Starting new instance of %s', program_name)
    logger.debug('Raw configuration: %s', config)

    # Import gnupg if necessary
    use_gpg = False
    # There is a better solution for the following for sure
    for acc, acc_settings in config.get('accounts').items():
        if acc_settings.get('enabled', False) and 'password_enc' in acc_settings:
            use_gpg = True
            break
    if use_gpg:
        import gnupg
        gpg = gnupg.GPG(homedir=config.get('settings').get('gpg_homedir', '~/.gnupg/'),
                        use_agent=config.get('settings').get('gpg_use_agent', True),
                        binary=config.get('settings').get('gpg_binary', 'gpg2'))
        gpg.encoding = 'utf-8'

    # Initialize connection pools
    imap_pool = {}
    for acc, acc_settings in config.get('accounts').items():
        if not acc_settings.get('enabled', False):
            continue

        # Check whether we got a plaintext password
        acc_password = acc_settings.get('password')
        if not acc_password:
            # Switch to GPG-encrypted password
            enc_password = gpg.decrypt(acc_settings.get('password_enc'))
            if not enc_password.ok:
                logger.error('%s: Failed to decrypt GPG message: %s', acc_settings.get('username'), enc_password.status)
                logger.debug('%s: GPG error: %s', acc_settings.get('username'), enc_password.stderr)
                exit(1)
            acc_password = str(enc_password)

        imap_pool[acc] = IMAP(logger=logger,
                              server=acc_settings.get('server'),
                              port=acc_settings.get('port'),
                              username=acc_settings.get('username'),
                              password=acc_password,
                              test=test)
        imap_pool[acc].connect()  # TODO error handling

    while True:
        for acc, acc_settings in config.get('accounts').items():
            if not acc_settings.get('enabled', False):
                continue
            pre_inbox = acc_settings.get('pre_inbox', 'PreInbox')
            pre_inbox_search = acc_settings.get('pre_inbox_search', 'ALL')
            sort_mailbox = acc_settings.get('sort_mailbox', None)

            mail_uids = imap_pool[acc].search_mails(pre_inbox, pre_inbox_search)
            if not mail_uids:
                logger.debug('%s: No mails found, continue with next mail account..', acc_settings.get('username'))
                continue

            mails = imap_pool[acc].fetch_mails(uids=mail_uids, mailbox=pre_inbox)
            for uid, mail in mails.items():
                match = False
                for filter_name, filter_rulesets in sorted(config.get('filters').get(acc).items()):
                    set_commands = filter_rulesets.get('commands', None)
                    set_rules = filter_rulesets.get('rules', None)
                    ruleset = RuleSet(logger=logger,
                                      name=filter_name,
                                      ruleset=set_rules,
                                      commands=set_commands,
                                      imap=imap_pool[acc],
                                      mail=(uid, mail),
                                      mailbox=pre_inbox)
                    match = ruleset.process()
                    if match:
                        break
                if not match and not sort_mailbox:
                    imap_pool[acc].set_mailflags([uid], pre_inbox, acc_settings.get('unmatched_mail_flags', ['\FLAGGED']))

            if sort_mailbox:
                logger.info('%s: Searching for mails that did not match any filter and moving them to %s', acc_settings.get('username'),
                            sort_mailbox)
                uids = imap_pool[acc].search_mails(pre_inbox)
                mails = imap_pool[acc].fetch_mails(uids=uids, mailbox=pre_inbox)
                for mail in mails:
                    imap_pool[acc].move_mail(mails[mail], pre_inbox, sort_mailbox, set_flags=[])

        break

    logger.debug('Shutting down tabellarius instance..')


if __name__ == '__main__':
    exit(main())



# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et

import os
import re
import logging
import logging.config
import yaml


class ConfigParser(object):
    def __init__(self, confdir, config=None):
        self.confdir = confdir

        if not config:
            config = {'settings': {}, 'accounts': {}, 'filters': {}}
        self.config = config

        self.parse_directory()

    def parse_directory(self):
        for dirname, subdirectories, files in os.walk(self.confdir):
            for file_name in files:
                file_path = '{0}/{1}'.format(dirname, file_name)
                if file_name.endswith('.yaml'):
                    with open(file_path, 'r') as stream:
                        data = yaml.load(stream)
                    if data:
                        for root, value in data.items():
                            if root == 'settings':
                                self.config[root] = value
                            elif root == 'accounts':
                                for account, settings in value.items():
                                    if account not in self.config[root].keys():
                                        self.config[root][account] = settings
                                    else:
                                        self.config[root][account].update(settings)
                            elif root == 'filters':
                                for account, filter_set in value.items():
                                    for filterset_name, filterset_data in filter_set.items():
                                        if account not in self.config[root].keys():
                                            self.config[root][account] = {}
                                        self.config[root][account].update({filterset_name: filterset_data})

    def dump(self):
        return self.config


class Helper(object):
    """
    Contains helper functions
    """

    @staticmethod
    def check_match(string, pattern):
        if string is None or len(string) == 0:
            return False

        # Basic match
        if pattern in string:
            return True

        # RegEx match
        pattern_re = re.compile(pattern)
        if pattern_re.match(string):
            return True

        return False

    @staticmethod
    def clean_field_names(field):
        if field[-1:] == '!':
            return (field, field[0:-1], True)
        else:
            return (field, field, False)

    @staticmethod
    def create_logger(program_name, config=None):
        if not config:
            config = {'version': 1}
        logger = logging.getLogger(program_name)
        logging.config.dictConfig(config)
        return logger



# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et

from misc import Helper


class RuleSet(object):
    def __init__(self, logger, name, mailbox, mail, ruleset=[], commands=[], imap=None):
        self.logger = logger
        self.name = name
        self.mailbox = mailbox
        self.mail = mail
        self.ruleset = ruleset
        self.commands = commands
        self.imap = imap
        self.supported_rule_operators = ['or', 'and']

    def process(self):
        match = False
        uid, mail = self.mail
        self.logger.debug('Checking whether mail uid="%s"; message-id="%s"; subject="%s" matches to ruleset %s', uid,
                          mail.get('message-id'), mail.get('subject'), self.name)
        match = self.parse_ruleset(mail, self.ruleset)
        if match:
            self.logger.debug('Ruleset matches!')
            self.apply_commands(uid, mail, self.mailbox)

        return match

    def apply_commands(self, uid, mail, mailbox):
        self.logger.debug('Applying commands for mail message-id="%s" of ruleset %s', mail.get('message-id'), self.name)

        for command in self.commands:
            cmd_type = command.get('type')
            cmd_flags_set = command.get('all_flags', [])
            #cmd_flags_delete = command.get('delete_flags', [])
            #cmd_flags_add = command.get('add_flags', [])

            if cmd_type == 'move':
                cmd_target = command.get('target')
                self.imap.move_mail(mail, mailbox, cmd_target, set_flags=cmd_flags_set)

    def parse_ruleset(self, mail, ruleset):
        for rule in ruleset:
            operator = sorted(rule.keys())[0]
            conditions = rule.get(operator)
            match = self.filter_match(mail, operator, conditions)
            if match:
                return match

    def filter_match(self, mail, operator, conditions):
        #print('Check match: operator={0} conditions={1}'.format(operator, conditions))
        if operator == 'or':
            last_match = True
            for condition in conditions:
                field_original, field, invert = Helper().clean_field_names(sorted(condition.keys())[0])
                pattern = condition.get(field_original)

                if field in self.supported_rule_operators:
                    last_match = self.filter_match(mail=mail, operator=field, conditions=pattern)
                    if last_match:
                        return True
                else:
                    lines = mail.get(field)
                    if lines is None:
                        last_match = False
                        continue
                    if type(lines) is not list:
                        lines = [lines]
                    for line in lines:
                        last_match = Helper().check_match(line, pattern)  # TODO improve
                        if last_match:
                            break
                    if invert:
                        last_match = not last_match
                    if last_match:
                        return True
            return last_match
        elif operator == 'and':
            for condition in conditions:
                field_original, field, invert = Helper().clean_field_names(sorted(condition.keys())[0])
                pattern = condition.get(field_original)
                if field in self.supported_rule_operators:
                    last_match = self.filter_match(mail=mail, operator=field, conditions=pattern)
                    if not last_match:
                        return False
                else:
                    lines = mail.get(field)
                    #if lines is None:
                    #    continue
                    if type(lines) is not list:
                        lines = [lines]
                    for line in lines:
                        last_match = Helper().check_match(line, pattern)  # TODO improve
                        if not last_match:
                            break
                    if invert:
                        last_match = not last_match
                    if not last_match:
                        return False
            return True
        else:
            print('operator {0} is not supported yet'.format(operator))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from distutils.core import setup

setup(name='tabellarius',
      version='0.0.2',
      description='A mail-sorting tool that is less annoying',
      author='Arnold Bechtoldt',
      author_email='mail@arnoldbechtoldt.com',
      url='https://github.com/bechtoldt/tabellarius',
      packages=['tabellarius'], )
